import numpy as np
from P4J.generator import synthetic_light_curve_generator

def test_generator_draw():
    lc_generator = synthetic_light_curve_generator(T=100.0, N=100, rseed=0)
    lc_generator.set_model(f0=1.23456, A=[1.0])
    mjd, mag, err = lc_generator.draw_noisy_time_series(SNR=20.0)
    assert mjd.shape == (100,)
    assert mag.shape == (100,)
    assert err.shape == (100,)
    assert np.allclose(lc_generator.get_fundamental_frequency(), 1.23456)




from __future__ import division, print_function
import numpy as np
from scipy.stats import gamma, exponnorm

class synthetic_light_curve_generator:
    def __init__(self, T, N, rseed=None):
        """
        Class for simple synthetic light curve generation
        
        Light curves are time series of stellar magnitude or flux as a 
        function of time. In the following we consider Earth-based surveys.
        Light curves are irregularly sampled because the measurements are 
        not taken at the same time every night due to observation 
        constraints. Light curves may have data gaps, i.e. periods of time 
        where no observations were registered. The noise present in the 
        light curves comes from the sky background, the Earth atmosphere, 
        telescope systematics, etc.
        
        Parameters
        ----------
        T: float
            Time span of the vector, i.e. how long it is in time
        N: positive integer
            Number of samples of the resulting time vector
        rseed: 
            Random seed to feed the random number generator
        
        """
        self.rseed = rseed
        self.t = irregular_sampling(T, N, rseed)
        self.A = 1.0
        
    def set_model(self, f0, A, model='trigonometric'):
        """
        See also
        --------
        trigonometric_model
        
        """
        self.f0 = f0
        self.y_clean, vary = trigonometric_model(self.t, f0, A)
        # y_clean is normalized to have unit standard deviation
        self.y_clean = self.y_clean/np.sqrt(vary) 
    
    def get_fundamental_frequency(self):
        return self.f0

    def get_clean_signal(self):
        return self.t, self.A*self.y_clean
        
    def draw_noisy_time_series(self, SNR=1.0, red_noise_ratio=0.25, outlier_ratio=0.0):
        """
        A function to draw a noisy time series based on the clean model 
        such that 
            y_noisy = y + yw + yr,
        where yw is white noise, yr is red noise and y will be rescaled 
        so that y_noisy complies with the specified signal-to-noise 
        ratio (SNR). 
        
        Parameters
        ---------
        SNR: float
            Signal-to-noise ratio of the resulting contaminated signal in
            decibels [dB]. SNR is defined as 
                SNR = 10*log(var_signal/var_noise), hence
            NR var_signal/var_noise
                 10  10
                  7   5
                  3   2
                  0   1
                 -3   0.5
                 -7   0.2
                -10   0.1
        red_noise_variance: float in [0, 1]
            The variance of the red noise component is set according to 
            Var(yw)*red_noise_ratio. Set this to zero to obtain uncertainties
            that explain the noise perfectly
        outlier_ratio: float in [0, 1]
            Percentage of outlier data points
            
        Returns
        -------
        t: ndarray
            Vector containing the time instants
        y_noisy: ndarray
            Vector containing the contaminated signal
        s: ndarray
            Vector containing the uncertainties associated to the white 
            noise component

        """
        if outlier_ratio < 0.0 or outlier_ratio > 1.0:
            raise ValueError("Outlier ratio must be in [0, 1]")
        if red_noise_ratio < 0.0:
            raise ValueError("Red noise ratio must be positive")
        np.random.seed(self.rseed)
        t = self.t
        y_clean = self.y_clean
        N = len(t)
        # First we generate s 
        s, mean_s_squared = generate_uncertainties(N, rseed=self.rseed)
        #print(mean_s_squared)
        #print(np.mean(s**2))
        # Draw a heteroscedastic white noise vector
        white_noise = np.random.multivariate_normal(np.zeros(N,), np.diag(s**2))
        # Now we generate a colored noise vector which is unaccounted by s
        red_noise_variance = mean_s_squared*red_noise_ratio
        # First order markovian process to generate 
        red_noise = first_order_markov_process(t, red_noise_variance, 1.0, rseed=self.rseed)
        
        # The following is not ok for irregularly sampled time series because
        # it assumes constant dt=1
        #phi=0.5
        #red_noise = np.random.randn(N)*np.sqrt(red_noise_variance)
        #for i in range(1, N):
        #    red_noise[i] = phi*red_noise[i-1] + np.sqrt(1 - phi**2)*red_noise[i]
        
        # The final noise vector
        #print("%f %f" % (np.var(white_noise)*red_noise_ratio, np.var(red_noise)))
        noise = white_noise + red_noise 
        var_noise = mean_s_squared + red_noise_variance
        SNR_unitless = 10.0**(SNR/10.0)
        self.A = np.sqrt(SNR_unitless*var_noise)
        y = self.A*y_clean
        y_noisy = y + noise
        # Add outliers with a certain percentage
        rperm = np.where(np.random.uniform(size=N) < outlier_ratio)[0]    
        outlier = np.random.uniform(5.0*np.std(y), 10.0*np.std(y), size=len(rperm))
        y_noisy[rperm] += outlier
        return t, y_noisy, s




def irregular_sampling(T, N, rseed=None):
    """
    Generates an irregularly sampled time vector by perturbating a 
    linearly spaced vector and latter deleting a certain number of 
    points
    
    Parameters
    ----------
    T: float
        Time span of the vector, i.e. how long it is in time
    N: positive integer
        Number of samples of the resulting time vector
    rseed: 
        Random seed to feed the random number generator
        
    Returns
    -------
    t_irr: ndarray
        An irregulary sampled time vector
        
    """
    sampling_period = (T/float(N))
    N = int(N)
    np.random.seed(rseed)    
    t = np.linspace(0, T, num=5*N)
    # First we add jitter
    t[1:-1] += sampling_period*0.5*np.random.randn(5*N-2)
    # Then we do a random permutation and keep only N points 
    P = np.random.permutation(5*N)
    t_irr = np.sort(t[P[:N]])
    return t_irr
    

def trigonometric_model(t, f0, A):
    """
    Generates a simple trigonometric model based on a sum of sine waves 
    
    Parameters
    ----------
    t: ndarray
        Vector containing the time instants where the model will be sampled
    f0: float
        Fundamental frequency of the model
    A: ndarray 
        Vector containing the amplitudes of the harmonics in the model. 
        The lenght of A defines the number of harmonics
        
    Returns
    -------
    y: ndarray
        Trigonometric model based
    var_y: float
        Analytical variance of the trigonometric model
        
    Example
    -------
    Assumming a time vector t has already been specified
    
    >>> import P4J
    >>> y_clean = P4J.trigonometric_model(t, f0=2.0, A=[1.0, 0.5, .25])
    
    """
    y = 0.0
    M = len(A)
    var_y = 0.0
    for k in range(0, M):
        y += A[k]*np.sin(2.0*np.pi*t*f0*(k+1))
        var_y += 0.5*A[k]**2
    return y, var_y


def first_order_markov_process(t, variance, time_scale, rseed=None):
    """
    Generates a correlated noise vector using a multivariate normal
    random number generator with zero mean and covariance
    
        Sigma_ij = s^2 exp(-|t_i - t_j|/l),
    
    where s is the variance and l is the time scale. 
    The Power spectral density associated to  this covariance is 
    
        S(f) = 2*l*s^2/(4*pi^2*f^2*l^2 +1),
        
    red noise spectrum is defined as proportional to 1/f^2. This 
    covariance is the one expected from a first order markov process 
    (Reference?)
    
    Parameters
    ---------
    t: ndarray
        A time vector for which the red noise vector will be sampled
    variance: positive float
        variance of the resulting red noise vector
    time_scale: positive float
        Parameter of the covariance matrix
        
    Returns
    -------
    red_noise: ndarray
        Vector containing the red noise realizations
        
    See also
    --------
    power_law_noise
        
    """
    if variance < 0.0:
        raise ValueError("Variance must be positive")
    if time_scale < 0.0:
        raise ValueError("Time scale must be positive")
    np.random.seed(rseed)
    N = len(t)
    mu = np.zeros(shape=(N,))
    if variance == 0.0:
        return mu
    dt = np.repeat(np.reshape(t, (1, -1)), N, axis=0)
    dt = np.absolute(dt - dt.T)  # This is NxN
    S = variance*np.exp(-np.absolute(dt)/time_scale)
    red_noise = np.random.multivariate_normal(mu, S)
    return red_noise


def power_law_noise(t, variance):
    """
    CAUTION: NOT THOROUGHLY TESTED
    Generates a red noise vector by first generating a covariance 
    function based on the expected red noise spectra. Red noise spectrum
    follows a power law
    
        S(f) = A*f^(-2),
    
    where A is a constant. Pink noise can be obtained if the exponent is
    changed to -1, and white noise if it is changed to 0.
    
    This Power Spectral density (PSD) is real and even then
    
        r(tau) = AT sum cos(2pi k tau/T)/k**2,
    
    where we assume f[k] = k*df, df = 1/T, Fs = N/T.
    
    The basel problem gives us: sum (1/k**2) approx pi**2/6, hence we can
    set c according to the desired variance at lag=0
     
    After the covariance is computed we can draw from a multivariate
    normal distribution to obtain a noise vector
    
    Parameters
    ---------
    t: ndarray
        A time vector for which the red noise vector will be sampled
    variance: positive float
        variance of the resulting red noise vector
        
    Returns
    -------
    red_noise: ndarray
        Vector containing the red noise realizations
        
    See also
    --------
    first_order_markov_process
    
    """
    if variance < 0.0:
        raise ValueError("Variance must be positive")
    mu = np.zeros(shape=(N,))
    if variance == 0.0:
        return mu
    N = len(t)    
    T = (t[-1] - t[0])
    c = (6.0*variance)/(T*np.pi**2)
    f = np.arange(1.0/T, 0.5*N/T, step=1.0/T)  # We ommit f=0.0
    k = f*T
    dt = np.repeat(np.reshape(t, (1, -1)), N, axis=0)
    dt = np.absolute(dt - dt.T)  # This is NxN
    S = np.zeros(shape=(N, N))
    for i in range(0, N):
        for j in range(i, N):
            S[i, j] = np.sum(np.cos(2.0*np.pi*k*dt[i, j]/T)/k**2)*T*c
            S[j, i] = S[i, j] 
    red_noise = np.random.multivariate_normal(mu, S)
    return red_noise


def generate_uncertainties(N, dist='Gamma', rseed=None):
    """
    This function generates a uncertainties for the white noise component
    in the synthetic light curve. 
    
    Parameters
    ---------
    N: positive integer
        Lenght of the returned uncertainty vector
    dist: {'EMG', 'Gamma'}
        Probability density function (PDF) used to generate the 
        uncertainties
    rseed:
        Seed for the random number generator
        
    Returns
    -------
    s: ndarray
        Vector containing the uncertainties
    expected_s_2: float
        Expectation of the square of s computed analytically
        
    """
    np.random.seed(rseed)  
    #print(dist)
    if dist == 'EMG':  # Exponential modified Gaussian
        # the mean of a EMG rv is mu + 1/(K*sigma)
        # the variance of a EMG rv is sigma**2 + 1/(K*sigma)**2
        K = 1.824328605481941
        sigma = 0.05*0.068768312946785953
        mu = 0.05*0.87452567616276777
        # IMPORTANT NOTE
        # These parameters were obtained after fitting uncertainties
        # coming from 10,000 light curves of the VVV survey
        expected_s_2 = sigma**2 + mu**2 + 2*K*mu*sigma + 2*K**2*sigma**2 
        s = exponnorm.rvs(K, loc=mu, scale=sigma, size=N)
    elif dist == 'Gamma':
        # The mean of a gamma rv is k*sigma
        # The variance of a gamma rv is k*sigma**2
        k = 3.0
        sigma = 0.05/k  #  mean=0.05, var=0.05**2/k
        s = gamma.rvs(k, loc=0.0, scale=sigma, size=N)
        expected_s_2 = k*(1+k)*sigma**2  
    return s, expected_s_2



#!/usr/bin/python
#cython: initializedcheck=False, boundscheck=False, wraparound=False, cdivision=True, profile=False

cimport cython
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from .utilities cimport argsort, unbiased_weighted_variance

ctypedef float DTYPE_t
ctypedef int ITYPE_t

cdef extern from "math.h":
    DTYPE_t sqrtf(DTYPE_t)
    DTYPE_t powf(DTYPE_t, DTYPE_t)
    DTYPE_t fmodf(DTYPE_t, DTYPE_t)

"""

Lafler Kinman String/Rope Length implementation
Ref: http://www.aanda.org/articles/aa/full/2002/17/aa2208/aa2208.html

The original method is modified to include uncertainties, the variance in
the denominator is replaced by the unbiased weighted variance, and the euclidean 
distance between sorted samples is replaced by a weighted version, i.e.
 sum[m[1:N] - m[0:N-1)**2/(e[1:N]**2 + e[0:N-1]**2)]/sum[1.0/(e[1:N]**2 + e[0:N-1]**2)]

Please cite the paper above if using this code
"""

cdef class LKSL:
    cdef Py_ssize_t N
    cdef DTYPE_t* phase
    cdef DTYPE_t* mjd
    cdef DTYPE_t* mag
    cdef DTYPE_t* err2
    cdef DTYPE_t normalizer
    cdef ITYPE_t* sorted_idx
    def __init__(self, DTYPE_t [::1] mjd, DTYPE_t [::1] mag, DTYPE_t [::1] err):
        cdef Py_ssize_t i, j, mat_idx
        self.N = mag.shape[0]
        self.mjd = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        self.mag = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        self.err2 = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        self.phase = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        self.sorted_idx = <ITYPE_t*>PyMem_Malloc(self.N*sizeof(ITYPE_t))
        if not self.mjd:
            raise MemoryError()
        if not self.phase:
            raise MemoryError()
        if not self.mag:
            raise MemoryError()
        if not self.err2:
            raise MemoryError()
        if not self.sorted_idx:
            raise MemoryError()
        for i in range(self.N):
            self.mjd[i] = mjd[i]
            self.mag[i] = mag[i]
            self.err2[i] = powf(err[i], 2.0)
        self.normalizer = 1.0/unbiased_weighted_variance(self.mag, self.err2, self.N)

    def eval_frequency(self, DTYPE_t freq):
        cdef Py_ssize_t i, j
        for i in range(self.N):
            self.phase[i] = fmodf(self.mjd[i], 1.0/freq)*freq  # output in [0.0, 1.0]
        argsort(self.phase, self.sorted_idx, self.N)
        cdef DTYPE_t err2_err2 = self.err2[self.sorted_idx[0]] + self.err2[self.sorted_idx[self.N-1]]
        cdef DTYPE_t err2_acum = 1.0/err2_err2
        cdef DTYPE_t SL = powf(self.mag[self.sorted_idx[0]] - self.mag[self.sorted_idx[self.N-1]], 2.0)/err2_err2
        for i in range(1, self.N):
            err2_err2 = self.err2[self.sorted_idx[i-1]] + self.err2[self.sorted_idx[i]]
            err2_acum += 1.0/err2_err2
            SL += powf(self.mag[self.sorted_idx[i-1]] - self.mag[self.sorted_idx[i]], 2.0)/err2_err2
        return 0.5*SL*self.normalizer/err2_acum

    def __dealloc__(self):
        PyMem_Free(self.mjd)
        PyMem_Free(self.mag)
        PyMem_Free(self.err2)
        PyMem_Free(self.phase)
        PyMem_Free(self.sorted_idx)





#!/usr/bin/python
#cython: initializedcheck=False, boundscheck=False, wraparound=False, cdivision=True, profile=False

cimport cython
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from .utilities cimport argsort, unbiased_weighted_variance

ctypedef float DTYPE_t
ctypedef int ITYPE_t

cdef extern from "math.h":
    DTYPE_t sqrtf(DTYPE_t)
    DTYPE_t powf(DTYPE_t, DTYPE_t)
    DTYPE_t floorf(DTYPE_t)
    DTYPE_t fmodf(DTYPE_t, DTYPE_t)

"""

Phase Dispersion Minimization
Ref: http://adsabs.harvard.edu/abs/1978ApJ...224..953S

This implementation follows the reference except that variance calculations
are replaced by an unbiased weighted variance to include uncertainties.

There is also the PDM2 method which fits splines to the binned data
http://www.stellingwerf.com/rfs-bin/index.cgi?action=PageView&id=29
"""
cdef class PDM:
    cdef Py_ssize_t N
    cdef DTYPE_t* phase
    cdef DTYPE_t* mjd
    cdef DTYPE_t* mag
    cdef DTYPE_t* err2
    cdef DTYPE_t normalizer
    cdef ITYPE_t* sorted_idx
    cdef ITYPE_t Nbins
    cdef DTYPE_t* tmp_mag
    cdef DTYPE_t* tmp_err2
    def __init__(self, DTYPE_t [::1] mjd, DTYPE_t [::1] mag, DTYPE_t [::1] err, ITYPE_t Nbins=10):
        cdef Py_ssize_t i, j, mat_idx
        self.Nbins = Nbins
        self.N = mag.shape[0]
        self.mjd = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        self.mag = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        self.err2 = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        self.phase = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        self.sorted_idx = <ITYPE_t*>PyMem_Malloc(self.N*sizeof(ITYPE_t))
        self.tmp_mag = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        self.tmp_err2 = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        if not self.mjd:
            raise MemoryError()
        if not self.phase:
            raise MemoryError()
        if not self.mag:
            raise MemoryError()
        if not self.err2:
            raise MemoryError()
        if not self.sorted_idx:
            raise MemoryError()
        if not self.tmp_mag:
            raise MemoryError()
        if not self.tmp_err2:
            raise MemoryError()
        for i in range(self.N):
            self.mjd[i] = mjd[i]
            self.mag[i] = mag[i]
            self.err2[i] = powf(err[i], 2.0)
        self.normalizer = 1.0/unbiased_weighted_variance(self.mag, self.err2, self.N)

    def eval_frequency(self, DTYPE_t freq):
        cdef Py_ssize_t i, j
        for i in range(self.N):
            self.phase[i] = fmodf(self.mjd[i], 1.0/freq)*freq  # output in [0.0, 1.0]
        # argsort(self.phase, self.sorted_idx, self.N)
        cdef DTYPE_t PDM_num=0.0, PDM_den=0.0
        cdef ITYPE_t samples_in_bin
        cdef DTYPE_t V1, V2
        for j in range(self.Nbins):
            samples_in_bin = 0
            V1 = V2 = 0.0
            for i in range(self.N):
                if floorf(self.phase[i]*self.Nbins) == j:
                    self.tmp_mag[samples_in_bin] = self.mag[i]
                    self.tmp_err2[samples_in_bin] = self.err2[i]
                    samples_in_bin += 1
                    V1 += 1.0/self.err2[i]
                    V2 += 1.0/powf(self.err2[i], 2.0)
            if samples_in_bin > 2:
                PDM_num += unbiased_weighted_variance(self.tmp_mag, self.tmp_err2, samples_in_bin)*(V1 - V2/V1)
                PDM_den += (V1 - V2/V1)

        return PDM_num*self.normalizer/PDM_den

    def __dealloc__(self):
        PyMem_Free(self.mjd)
        PyMem_Free(self.mag)
        PyMem_Free(self.err2)
        PyMem_Free(self.phase)
        PyMem_Free(self.sorted_idx)
        PyMem_Free(self.tmp_mag)
        PyMem_Free(self.tmp_err2)





import numpy as np
from numpy.testing import assert_allclose
from P4J import periodogram
from P4J.generator import synthetic_light_curve_generator

def create_test_data():
    lc_generator = synthetic_light_curve_generator(T=100.0, N=50, rseed=0)
    lc_generator.set_model(f0=1.23456, A=[1.0, 0.5, 0.25])
    mjd, mag, err = lc_generator.draw_noisy_time_series(SNR=2.0)
    return mjd, mag, err

def test_standardize(tol=1e-7):
    my_per = periodogram(method='AOV')
    mjd, mag, err = create_test_data()
    my_per.set_data(mjd, mag, err, standardize=True)
    assert_allclose(my_per.mag[:3], np.array([-0.87692106, 0.44308114, 0.94811785], dtype=np.float32), atol=tol)
    assert_allclose(my_per.err[:3], np.array([1.399464   , 0.70398176, 1.4657015], dtype=np.float32), atol=tol)
    
def test_removenan():
    my_per = periodogram(method='AOV')
    mjd, mag, err = create_test_data()
    mjd[1], mag[5], err[8] = np.nan, np.nan, np.nan
    my_per.set_data(mjd, mag, err, remove_nan=True)
    assert len(my_per.mag) == len(mjd)-3    

def fit_eval_periodogram(method):
    my_per = periodogram(method=method)
    mjd, mag, err = create_test_data()
    my_per.set_data(mjd, mag, err)
    my_per.frequency_grid_evaluation(fmin=0.01, fmax=10., fresolution=1e-4)
    my_per.finetune_best_frequencies(n_local_optima=3, fresolution=1e-5)
    return my_per.get_best_frequencies()

def test_periodogram_mhaov():
    best_freq, best_per = fit_eval_periodogram('MHAOV')

    assert best_freq.dtype == np.float32
    assert best_per.dtype == np.float32
    assert best_freq.shape == (3,)
    assert best_per.shape == (3,)
    assert_allclose(best_freq, np.array([1.234198 , 9.704909, 6.981068], dtype=np.float32))
    assert_allclose(best_per, np.array([58.99221, 40.944813, 37.43901], dtype=np.float32))
    
def test_periodogram_aov():
    best_freq, best_per = fit_eval_periodogram('AOV')

    assert best_freq.dtype == np.float32
    assert best_per.dtype == np.float32
    assert best_freq.shape == (3,)
    assert best_per.shape == (3,)
    assert_allclose(best_freq, np.array([1.2353979, 4.0710936, 9.823684], dtype=np.float32))
    assert_allclose(best_per, np.array([9.8947315, 8.2189865, 8.14342], dtype=np.float32))
    
def test_periodogram_pdm():
    best_freq, best_per = fit_eval_periodogram('PDM1')

    assert best_freq.dtype == np.float32
    assert best_per.dtype == np.float32
    assert best_freq.shape == (3,)
    assert best_per.shape == (3,)
    assert_allclose(best_freq, np.array([1.2352979, 1.235708, 1.234098], dtype=np.float32))
    assert_allclose(best_per, np.array([-0.19907248, -0.23332588, -0.28197145], dtype=np.float32))

def test_periodogram_qmi():
    best_freq, best_per = fit_eval_periodogram('QMIEU')
    
    assert best_freq.dtype == np.float32
    assert best_per.dtype == np.float32
    assert best_freq.shape == (3,)
    assert best_per.shape == (3,)
    assert_allclose(best_freq, np.array([1.234878, 6.06321, 3.4890144], dtype=np.float32))
    assert_allclose(best_per, np.array([0.03774722, 0.029475132, 0.028804852], dtype=np.float32))






#!/usr/bin/python
#cython: initializedcheck=False, boundscheck=False, wraparound=False, cdivision=True, profile=False

cimport cython
from cython.operator cimport dereference
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from .utilities cimport argsort, mean, weighted_mean, unbiased_weighted_variance

ctypedef float DTYPE_t
ctypedef int ITYPE_t

cdef extern from "math.h":
    DTYPE_t sqrtf(DTYPE_t)
    DTYPE_t powf(DTYPE_t, DTYPE_t)
    DTYPE_t floorf(DTYPE_t)
    DTYPE_t fmodf(DTYPE_t, DTYPE_t)

"""

Analysis of Variance periodograma
Ref: http://adsabs.harvard.edu/full/1989MNRAS.241..153S

Please cite the paper above if using this code
"""

cdef class AOV:
    cdef Py_ssize_t N
    cdef DTYPE_t* phase
    cdef DTYPE_t* mjd
    cdef DTYPE_t* mag
    cdef DTYPE_t* err
    cdef DTYPE_t normalizer, barx
    cdef ITYPE_t Nbins
    cdef DTYPE_t* tmp_mag
    cdef DTYPE_t* tmp_err
    cdef ITYPE_t* tmp_sizes
    cdef int use_errorbars

    def __init__(self, DTYPE_t [::1] mjd, DTYPE_t [::1] mag, DTYPE_t [::1] err, ITYPE_t Nbins=10, use_errorbars=1):
        cdef Py_ssize_t i, j, mat_idx
        self.Nbins = Nbins
        self.use_errorbars = use_errorbars
        self.N = mag.shape[0]
        self.mjd = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        self.mag = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        self.err = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        self.phase = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        #self.tmp_mag = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        #self.tmp_err = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        self.tmp_mag = <DTYPE_t*>PyMem_Malloc(self.N*self.Nbins*sizeof(DTYPE_t))
        self.tmp_err = <DTYPE_t*>PyMem_Malloc(self.N*self.Nbins*sizeof(DTYPE_t))
        self.tmp_sizes = <ITYPE_t*>PyMem_Malloc(self.Nbins*sizeof(ITYPE_t))
        if not self.mjd:
            raise MemoryError()
        if not self.phase:
            raise MemoryError()
        if not self.mag:
            raise MemoryError()
        if not self.err:
            raise MemoryError()
        if not self.tmp_mag:
            raise MemoryError()
        if not self.tmp_err:
            raise MemoryError()
        if not self.tmp_sizes:
            raise MemoryError()
        for i in range(self.N):
            self.mjd[i] = mjd[i]
            self.mag[i] = mag[i]
            self.err[i] = err[i]
        self.normalizer = (self.N - Nbins)/(Nbins-1)
        if self.use_errorbars:
            self.barx = weighted_mean(self.mag, self.err, self.N)
        else:
            self.barx = mean(self.mag, self.N)
        

    def eval_frequency(self, DTYPE_t freq):
        # More memory consuming but almost twice as fast!
        cdef Py_ssize_t i, j
        for i in range(self.N):
            self.phase[i] = fmodf(self.mjd[i], 1.0/freq)*freq  # output in [0.0, 1.0]
        cdef DTYPE_t num=0.0, den=0.0
        cdef ITYPE_t samples_in_bin
        cdef DTYPE_t barxi
        
        for j in range(self.Nbins):
            self.tmp_sizes[j] = 0
            
        for i in range(self.N):
            j = (int)(floorf(self.phase[i]*self.Nbins))
            self.tmp_mag[self.tmp_sizes[j] + self.N*j] = self.mag[i]
            self.tmp_err[self.tmp_sizes[j] + self.N*j] = self.err[i]
            self.tmp_sizes[j] += 1 
            
        for j in range(self.Nbins):
            if self.tmp_sizes[j] > 1:
                if self.use_errorbars:
                    barxi = weighted_mean((self.tmp_mag + self.N*j), 
                                          (self.tmp_err + self.N*j),
                                          self.tmp_sizes[j])
                else:
                    barxi = mean((self.tmp_mag + self.N*j), self.tmp_sizes[j])
                for i in range(self.tmp_sizes[j]):
                    den += (self.tmp_mag[i + self.N*j] - barxi)**2
                num += self.tmp_sizes[j]*(barxi - self.barx)**2
                
        return self.normalizer*num/den

    
    def eval_frequency_old(self, DTYPE_t freq):
        cdef Py_ssize_t i, j
        for i in range(self.N):
            self.phase[i] = fmodf(self.mjd[i], 1.0/freq)*freq  # output in [0.0, 1.0]
        cdef DTYPE_t num=0.0, den=0.0 # These are the unormalized s1 and s2 in the paper
        cdef ITYPE_t samples_in_bin
        cdef DTYPE_t barxi
        for j in range(self.Nbins):
            samples_in_bin = 0
            for i in range(self.N):
                if floorf(self.phase[i]*self.Nbins) == j:
                    self.tmp_mag[samples_in_bin] = self.mag[i]
                    self.tmp_err[samples_in_bin] = self.err[i]
                    samples_in_bin += 1

            if samples_in_bin > 1:
                if self.use_errorbars:
                    barxi = weighted_mean(self.tmp_mag, self.tmp_err, samples_in_bin)
                else:
                    barxi = mean(self.tmp_mag, samples_in_bin)
                for i in range(samples_in_bin):
                    den += (self.tmp_mag[i] - barxi)**2
                num += samples_in_bin*(barxi - self.barx)**2
                
        return self.normalizer*num/den

    def __dealloc__(self):
        PyMem_Free(self.mjd)
        PyMem_Free(self.mag)
        PyMem_Free(self.err)
        PyMem_Free(self.phase)
        PyMem_Free(self.tmp_mag)
        PyMem_Free(self.tmp_err)
        PyMem_Free(self.tmp_sizes)





#!/usr/bin/python
#cython: initializedcheck=False, boundscheck=False, wraparound=False, cdivision=True, profile=False

cimport cython
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from .utilities cimport weighted_mean

ctypedef float DTYPE_t
ctypedef int ITYPE_t

cdef extern from "math.h":
    DTYPE_t sqrtf(DTYPE_t)
    DTYPE_t powf(DTYPE_t, DTYPE_t)
    DTYPE_t floorf(DTYPE_t)
    DTYPE_t cosf(DTYPE_t)
    DTYPE_t sinf(DTYPE_t)


"""

Multiharmonic weighted AoV Periodogram
Ref: http://adsabs.harvard.edu/abs/1996ApJ...460L.107S

This implementation follows the c code found at
http://users.camk.edu.pl/alex/soft/aovgui.tgz

Please cite the paper above if using this code
"""

cdef DTYPE_t M_PI = 3.1415926535897

cdef DTYPE_t* allocate_and_verify(Py_ssize_t N):
    cdef DTYPE_t* array = <DTYPE_t*>PyMem_Malloc(N*sizeof(DTYPE_t))
    if not array:
        raise MemoryError()
    return array

cdef class MHAOV:
    cdef Py_ssize_t N
    cdef DTYPE_t* mjd
    cdef DTYPE_t* mag
    cdef DTYPE_t* err
    cdef ITYPE_t Nharmonics
    cdef ITYPE_t mode
    cdef DTYPE_t d1, d2
    cdef public DTYPE_t wmean, wvar
    cdef DTYPE_t* zr 
    cdef DTYPE_t* zi
    cdef DTYPE_t* znr
    cdef DTYPE_t* zni
    cdef DTYPE_t* pr
    cdef DTYPE_t* pi
    cdef DTYPE_t* cfr
    cdef DTYPE_t* cfi

    def __init__(self, DTYPE_t [::1] mjd, DTYPE_t [::1] mag, DTYPE_t [::1] err, ITYPE_t Nharmonics=1, ITYPE_t mode=1):
        cdef Py_ssize_t i
        if Nharmonics < 1:
            raise ValueError("Number of harmonics has to be greater or equal to 1")
        self.Nharmonics = Nharmonics
        self.N = mag.shape[0]
        self.mjd = allocate_and_verify(self.N)
        self.mag = allocate_and_verify(self.N)
        self.err = allocate_and_verify(self.N)
        self.zr = allocate_and_verify(self.N)
        self.zi = allocate_and_verify(self.N)
        self.znr = allocate_and_verify(self.N)
        self.zni = allocate_and_verify(self.N)
        self.pr = allocate_and_verify(self.N)
        self.pi = allocate_and_verify(self.N)
        self.cfr = allocate_and_verify(self.N)
        self.cfi = allocate_and_verify(self.N)
        
        for i in range(self.N):
            self.mjd[i] = mjd[i]
            self.mag[i] = mag[i]
            self.err[i] = err[i]
        
        self.wmean = weighted_mean(self.mag, self.err, self.N)
        self.d1 = Nharmonics*2.0
        self.d2 = self.N - Nharmonics*2 - 1
        self.wvar = 0.0
        for i in range(self.N):
            self.wvar += powf(self.mag[i] - self.wmean, 2.)/powf(self.err[i], 2.)
            
        self.mode = mode # 0: RAW, 1: F
        

    def eval_frequency(self, DTYPE_t freq):
        cdef Py_ssize_t i, j
        cdef DTYPE_t sn, alr, ali, scr, sci
        cdef DTYPE_t aov=0.0
        cdef DTYPE_t arg, tmp, sr, si
        for i in range(self.N):
            arg = self.mjd[i]*freq
            arg = 2.0*M_PI*(arg - floorf(arg))
            self.zr[i] = cosf(arg)
            self.zi[i] = sinf(arg)
            self.znr[i] = 1.
            self.pr[i] = 1./self.err[i]
            self.zni[i] = 0.
            self.pi[i] = 0.
            self.cfr[i] = (self.mag[i] - self.wmean)*cosf(self.Nharmonics*arg)/self.err[i]
            self.cfi[i] = (self.mag[i] - self.wmean)*sinf(self.Nharmonics*arg)/self.err[i]
        for j in range(2*self.Nharmonics+1):
            sn = alr = ali = scr = sci = 0.0
            for i in range(self.N):
                sn += self.pr[i]**2 + self.pi[i]**2
                alr += (self.zr[i]*self.pr[i] - self.zi[i]*self.pi[i])/self.err[i]
                ali += (self.zr[i]*self.pi[i] + self.zi[i]*self.pr[i])/self.err[i]
                scr += self.pr[i]*self.cfr[i] + self.pi[i]*self.cfi[i]
                sci += self.pr[i]*self.cfi[i] - self.pi[i]*self.cfr[i]
            sn = max(sn, 1e-9)
            alr = alr/sn
            ali = ali/sn
            aov += (scr**2 + sci**2)/sn
            for i in range(self.N):
                sr = alr*self.znr[i] - ali*self.zni[i]
                si = alr*self.zni[i] + ali*self.znr[i]
                tmp = self.pr[i]*self.zr[i] - self.pi[i]*self.zi[i] - sr*self.pr[i] - si*self.pi[i]
                self.pi[i] = self.pr[i]*self.zi[i] + self.pi[i]*self.zr[i] + sr*self.pi[i] - si*self.pr[i]
                self.pr[i] = tmp
                tmp = self.znr[i]*self.zr[i] - self.zni[i]*self.zi[i]
                self.zni[i] = self.zni[i]*self.zr[i] + self.znr[i]*self.zi[i]
                self.znr[i] = tmp
        if self.mode == 0:
            return aov
        elif self.mode == 1:        
            return (self.d2/self.d1)*aov/max(self.wvar - aov, 1e-9)
            
    def __dealloc__(self):
        PyMem_Free(self.mjd)
        PyMem_Free(self.mag)
        PyMem_Free(self.err)
        PyMem_Free(self.zr)
        PyMem_Free(self.zi)
        PyMem_Free(self.znr)
        PyMem_Free(self.zni)
        PyMem_Free(self.pr)
        PyMem_Free(self.pi)
        PyMem_Free(self.cfr)
        PyMem_Free(self.cfi)







ctypedef float DTYPE_t
ctypedef int ITYPE_t

cdef void argsort(DTYPE_t*, ITYPE_t*, Py_ssize_t)
cdef DTYPE_t mean(DTYPE_t*, Py_ssize_t)
cdef DTYPE_t weighted_mean(DTYPE_t*, DTYPE_t*, Py_ssize_t)
cdef DTYPE_t unbiased_weighted_variance(DTYPE_t*, DTYPE_t*, Py_ssize_t)




#!/usr/bin/python
#cython: initializedcheck=False, boundscheck=False, wraparound=False, cdivision=True, profile=False

cimport cython
from cpython.mem cimport PyMem_Malloc, PyMem_Free

cdef extern from "math.h":
    DTYPE_t powf(DTYPE_t, DTYPE_t)


"""
Computes the sample mean
"""
cdef DTYPE_t mean(DTYPE_t* data, Py_ssize_t N):
    cdef DTYPE_t acum = 0.0
    cdef Py_ssize_t i
    for i in range(N):
        acum += data[i]
    return acum/N

"""
Computes the weighted sample mean
"""
cdef DTYPE_t weighted_mean(DTYPE_t* data, DTYPE_t* err, Py_ssize_t N):
    cdef DTYPE_t w_mean = 0.0
    cdef DTYPE_t w_sum = 0.0
    cdef Py_ssize_t i
    for i in range(N):
        w_sum += 1.0/powf(err[i], 2.0)
        w_mean += data[i]/powf(err[i], 2.0)
    return w_mean/w_sum


"""
Computes an unbiased estimator of the weighted variance, where w_i = (1.0/e_i**2)
"""
cdef DTYPE_t unbiased_weighted_variance(DTYPE_t* data, DTYPE_t* err2, Py_ssize_t N):
    cdef DTYPE_t w_mean = 0.0
    cdef DTYPE_t w_var = 0.0
    cdef DTYPE_t V1 = 0.0, V2 = 0.0
    cdef Py_ssize_t i
    for i in range(N):
        V1 += 1.0/err2[i]
        V2 += 1.0/powf(err2[i], 2.0)
        w_mean += data[i]/err2[i]
    for i in range(N):
        w_var += powf(data[i] - w_mean/V1, 2.0)/err2[i]
    return w_var/(V1 - V2/V1)


"""

Argsort implementation
Credit: https://github.com/jcrudy/cython-argsort/blob/master/cyargsort/argsort.pyx

"""
cdef extern from "stdlib.h":
    ctypedef void const_void "const void"
    void qsort(void *base, int nmemb, int size,
            int(*compar)(const_void *, const_void *)) nogil

cdef struct Sorter:
    Py_ssize_t index
    DTYPE_t value

cdef int _compare(const_void *a, const_void *b) noexcept:
    cdef DTYPE_t v = ((<Sorter*>a)).value-((<Sorter*>b)).value
    if v < 0: return -1
    if v >= 0: return 1

cdef void cyargsort(DTYPE_t* data, Sorter * order, Py_ssize_t N):
    cdef Py_ssize_t i
    for i in range(N):
        order[i].index = i
        order[i].value = data[i]
    qsort(<void *> order, N, sizeof(Sorter), _compare)

cdef void argsort(DTYPE_t* data, ITYPE_t* order, Py_ssize_t N):
    cdef Py_ssize_t i
    cdef Sorter *order_struct = <Sorter *> PyMem_Malloc(N*sizeof(Sorter))
    cyargsort(data, order_struct, N)
    for i in range(N):
        order[i] = order_struct[i].index
    PyMem_Free(order_struct)




#!/usr/bin/python
# -*- coding: utf-8 -*-

from __future__ import division
import numpy as np



"""
    Computes a given quantile of the data considering
    that each sample has a weight. 
    x is a N float array
    weights is a N float array, it expects sum w_i = 1
    quantile is a float in [0.0, 1.0]
"""
def weighted_quantile(x, weights, quantile):
    I = np.argsort(x)
    sort_x = x[I]
    sort_w = weights[I]
    acum_w = np.add.accumulate(sort_w)
    norm_w = (acum_w - 0.5*sort_w)/acum_w[-1] 
    interpq = np.searchsorted(norm_w, [quantile])[0] 
    if interpq == 0:
        return sort_x[0]
    elif interpq == len(x):
        return sort_x[-1]
    else:
        tmp1 = (norm_w[interpq] - quantile)/(norm_w[interpq] - norm_w[interpq-1])
        tmp2 = (quantile - norm_w[interpq-1])/(norm_w[interpq] - norm_w[interpq-1])
        assert tmp1>=0 and tmp2>=0 and tmp1<=1 and tmp2<=1 
        return sort_x[interpq-1]*tmp1 + sort_x[interpq]*tmp2


"""
    Computes the weighted interquartile range (IQR) of x.
    x is a N float array
    weights is a N float array, it expects sum w_i = 1
"""

def wIQR(x, weights):
    return weighted_quantile(x, weights, 0.75) - weighted_quantile(x, weights, 0.25)

"""
    Computes the weighted standard deviation of x.
    x is a N float array
    weights is a N float array, it expects sum w_i = 1
"""
def wSTD(x, weights):
    wmean = np.average(x, weights=weights)
    return np.sqrt(np.average((x - wmean)**2, weights=weights))

"""
    Computes a robust measure of scale by comparing the weighted versions of the 
    standard deviation and the interquartile range of x.
    x is a N float array
    weights is a N float array, it expects sum w_i = 1
"""

def robust_scale(x, weights):
    return np.amin([wSTD(x, weights), wIQR(x, weights)/1.349])

def robust_loc(x, weights):
    return weighted_quantile(x, weights, 0.5)





import numpy as np
from numpy.testing import assert_allclose
from P4J import MultiBandPeriodogram
from P4J.generator import synthetic_light_curve_generator

def create_test_data():
    lc_generator = synthetic_light_curve_generator(T=100.0, N=50, rseed=0)
    lc_generator.set_model(f0=1.23456, A=[1.0, 0.5, 0.25])
    mjd1, mag1, err1 = lc_generator.draw_noisy_time_series(SNR=2.0)
    mjd2, mag2, err2 = lc_generator.draw_noisy_time_series(SNR=3.0)
    mjds = np.hstack((mjd1, mjd2))
    mags = np.hstack((mag1, mag2))
    errs = np.hstack((err1, err2))
    fids = np.array([['r']*50 + ['g']*50])[0, :]
    return mjds, mags, errs, fids


def test_mbperiodogram():
    my_per = MultiBandPeriodogram(method="MHAOV")
    mjds, mags, errs, fids = create_test_data()
    my_per.set_data(mjds, mags, errs, fids)
    my_per.frequency_grid_evaluation(fmin=0.01, fmax=10., fresolution=1e-4)
    my_per.finetune_best_frequencies(n_local_optima=3, fresolution=1e-5)
    best_freq, best_per = my_per.get_best_frequencies()
    assert len(my_per.per_single_band) == len(np.unique(fids))
    assert_allclose(best_freq, np.array([1.2341979, 9.704909, 0.86988866], dtype=np.float32))
    assert_allclose(best_per, np.array([125.610535,  81.329666,  74.506386], dtype=np.float32))




"""

P4J is a python package for period detection on irregularly sampled and
heteroscedastic time series based on Information Theoretic objective functions.
P4J was developed for astronomical light curves, irregularly sampled time
series of stellar magnitude or flux. The core of this package is a class called
periodogram that sweeps an array of periods/frequencies looking for the one
that maximizes a given criteria. The main contribution of this work is a
criterion for period detection based on the maximization of Cauchy-Schwarz
Quadratic Mutual Information [1]. Information theoretic criteria incorporate
information on the whole probability density function of the process and are
more robust than classical second-order statistics based criteria [2, 3, 4].
For comparison P4J also incorporates classical methods for period detection
used in astronomy such as Phase Dispersion Minimization [5], Lafler-Kinman's
string length [6], and the orthogonal multiharmonic AoV periodogram [7].

"""

__version__ = '1.1.2'

from .generator import synthetic_light_curve_generator
from .periodograms import periodogram, MultiBandPeriodogram



import os
from setuptools import setup, Extension, find_packages
from Cython.Build import cythonize
import numpy as np

include_dirs = ['.', np.get_include()]
library_dirs = []
if os.name == 'nt':  # Windows, assumming MSVC compiler
    libraries = []
    compiler_args = ['/Ox', '/fp:fast']
elif os.name == 'posix':  # UNIX, assumming GCC compiler
    libraries = ['m']
    compiler_args = ['-O3', '-ffast-math']

extensions = [
    Extension("*",
              sources=[os.path.join("src", "P4J", "algorithms", "*.pyx")],
              extra_compile_args=compiler_args,
              include_dirs=include_dirs,
              libraries=libraries,
              library_dirs=library_dirs
              )]

setup(
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    ext_modules=cythonize(extensions,
                          annotate=False,
                          language_level=3),
)



#!/usr/bin/python
#cython: initializedcheck=False, boundscheck=False, wraparound=False, cdivision=True, profile=False

cimport cython
#import numpy as np
#from numpy cimport ndarray, double_t, int_t
#from libc.math cimport sqrt, pow, M_PI, log
from cpython.mem cimport PyMem_Malloc, PyMem_Free
#from scipy.special.cython_special cimport i0

#DTYPE = np.double
ctypedef float DTYPE_t
#ctypedef int_t ITYPE_t

cdef extern from "math.h":
    DTYPE_t cosf(DTYPE_t)
    DTYPE_t fabsf(DTYPE_t)
    DTYPE_t sqrtf(DTYPE_t)
    DTYPE_t powf(DTYPE_t, DTYPE_t)
    DTYPE_t logf(DTYPE_t)
    DTYPE_t expf(DTYPE_t)
    DTYPE_t fmodf(DTYPE_t, DTYPE_t)

cdef DTYPE_t M_PI = 3.1415926535897

"""
Fills the Information Potential matrix using a different kernels

Wrapped kernels are meant for angle data.

"""

cdef inline void IP_wrappednormal(DTYPE_t* IP, DTYPE_t* angle, DTYPE_t h_KDE, Py_ssize_t N):
    cdef Py_ssize_t i, j
    cdef DTYPE_t distance
    cdef DTYPE_t denominator = sqrtf(2.0*M_PI*2.0)*h_KDE
    for i in range(N):
        IP[indexMatrixToVector(i, i, N)] = 1.0/denominator
        for j in range(i+1, N):
            # This is not quite right, it should be an infinite sum of |a_i - a_j - k2PI|
            distance = fabsf(angle[i]-angle[j])
            if distance > M_PI:
                distance -= 2.0*M_PI
            IP[indexMatrixToVector(i, j, N)] = expf(-0.25*powf(distance/h_KDE, 2))/denominator

"""
cdef inline void IP_vonmises(DTYPE_t* IP, DTYPE_t* angle, DTYPE_t h_KDE, Py_ssize_t N):
    cdef Py_ssize_t i, j
    cdef DTYPE_t kappa = 1.0/powf(2.0*M_PI*h_KDE, 2.0)
    cdef DTYPE_t denominator = 2.0*M_PI*i0(kappa)
    for i in range(N):
        IP[indexMatrixToVector(i, i, N)] = expf(kappa)/denominator
        for j in range(i+1, N):
            IP[indexMatrixToVector(i, j, N)] = expf(kappa*cosf(angle[i] - angle[j]))/denominator
"""

cdef inline void IP_wrappedcauchy(DTYPE_t* IP, DTYPE_t* angle, DTYPE_t h_KDE, Py_ssize_t N):
    cdef Py_ssize_t i, j
    cdef DTYPE_t rho = expf(-h_KDE)
    for i in range(N):
        IP[indexMatrixToVector(i, i, N)] = (1.+rho)/(2.*M_PI*(1.-rho))
        for j in range(i+1, N):
            IP[indexMatrixToVector(i, j, N)] = (1.-rho**2)/(2.*M_PI*(1.+rho**2-2.*rho*cosf(angle[i]-angle[j])))
 
cdef inline void IP_gaussian(DTYPE_t* IP, DTYPE_t [::1] data, DTYPE_t [::1] h_data, DTYPE_t h_KDE, Py_ssize_t N):
    cdef Py_ssize_t i, j
    cdef DTYPE_t gauss_var, delta2, h_KDE2 = 2.0*powf(h_KDE, 2.0)
    cdef DTYPE_t* h_data2 = <DTYPE_t*>PyMem_Malloc(N*sizeof(DTYPE_t))
    for i in range(N):
        h_data2[i] = powf(h_data[i], 2.0)
    for i in range(N):
        gauss_var = h_KDE2 + 2.0*h_data2[i]
        IP[indexMatrixToVector(i, i, N)] = 1.0/sqrtf(2.0*M_PI*gauss_var)
        for j in range(i+1, N):
            delta2 = powf(data[i] - data[j], 2.0)
            gauss_var = h_KDE2 + h_data2[i] + h_data2[j]
            IP[indexMatrixToVector(i, j, N)] = expf(-0.5*delta2/gauss_var)/sqrtf(2.0*M_PI*gauss_var)
    PyMem_Free(h_data2)


"""
 Testing Abramson's weighted KDE:
"""
cdef inline void IP_gaussian2(DTYPE_t* IP, DTYPE_t [::1] data, DTYPE_t [::1] h_data, DTYPE_t h_KDE, Py_ssize_t N):
    cdef Py_ssize_t i, j
    cdef DTYPE_t gauss_var, delta2, h_KDE2 = 2.0*powf(h_KDE, 2.0)
    cdef DTYPE_t* w = <DTYPE_t*>PyMem_Malloc(N*sizeof(DTYPE_t))
    cdef DTYPE_t sum_w = 0.0
    for i in range(N):
        w[i] = 1.0/powf(h_data[i], 2.0)
        sum_w += w[i]
    for i in range(N):
        w[i] = w[i]/sum_w
    for i in range(N):
        gauss_var = h_KDE2
        IP[indexMatrixToVector(i, i, N)] = powf(w[i], 2.0)/sqrtf(2.0*M_PI*gauss_var)
        for j in range(i+1, N):
            delta2 = powf(data[i] - data[j], 2.0)
            gauss_var = h_KDE2
            IP[indexMatrixToVector(i, j, N)] = w[i]*w[j]*expf(-0.5*delta2/gauss_var)/sqrtf(2.0*M_PI*gauss_var)
    PyMem_Free(w)


cdef inline void IP_cauchy(DTYPE_t* IP, DTYPE_t [::1] data, DTYPE_t [::1] err, DTYPE_t h_KDE, Py_ssize_t N):
    cdef Py_ssize_t i, j
    cdef DTYPE_t delta2
    cdef DTYPE_t* w = <DTYPE_t*>PyMem_Malloc(N*sizeof(DTYPE_t))
    cdef DTYPE_t w_sum = 0.0
    for i in range(N):
        w[i] = 1.0/powf(err[i], 2.0)
        w_sum += w[i]
    for i in range(N):
        w[i] = w[i]/w_sum
    for i in range(N):
        IP[indexMatrixToVector(i, i, N)] = powf(w[i], 2.0)/(M_PI*2.0*h_KDE)
        for j in range(i+1, N):
            delta2 = powf((data[i] - data[j])/(2.0*h_KDE), 2.0)
            IP[indexMatrixToVector(i, j, N)] = w[i]*w[j]/(M_PI*2.0*h_KDE*(1.0 + delta2))
    PyMem_Free(w)

cdef inline Py_ssize_t indexMatrixToVector(Py_ssize_t i, Py_ssize_t j, Py_ssize_t N):
    # Only works for i <= j, which is always the case here
    return i*N - (i-1)*i/2 + j - i

cdef class QMI:
    cdef DTYPE_t* IP_M
    cdef DTYPE_t* IP_P
    cdef Py_ssize_t N
    cdef DTYPE_t* VC1
    cdef DTYPE_t* VC2
    cdef DTYPE_t VM1
    cdef DTYPE_t h_KDE_P
    cdef DTYPE_t* angle
    cdef DTYPE_t* mjd
    cdef int mode # 0: Cauchy, 1: 
    def __init__(self, DTYPE_t [::1] mjd, DTYPE_t [::1] mag, DTYPE_t [::1] err, DTYPE_t h_KDE_M, DTYPE_t h_KDE_P, int mode=0, int kernel=0):
        cdef Py_ssize_t i, j, mat_idx
        self.mode = mode
        self.N = mag.shape[0]
        self.mjd = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        self.angle = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        if not self.mjd:
            raise MemoryError()
        if not self.angle:
            raise MemoryError()
        for i in range(self.N):
            self.mjd[i] = mjd[i]
        #self.h_KDE_P = 0.9*(1.0/sqrtf(12.0))*powf(self.N, -0.2)  # bandwidth considering a uniform distribution in phase
        self.h_KDE_P = h_KDE_P
        self.IP_M = <DTYPE_t*>PyMem_Malloc(self.N*(self.N+1)/2*sizeof(DTYPE_t))
        self.IP_P = <DTYPE_t*>PyMem_Malloc(self.N*(self.N+1)/2*sizeof(DTYPE_t))
        if not self.IP_M:
            raise MemoryError()
        if not self.IP_P:
            raise MemoryError()
        # Fill the IP matrix of the magnitudes
        if kernel == 0:
            IP_gaussian(self.IP_M, mag, err, h_KDE_M, self.N)
        elif kernel == 1:
            IP_cauchy(self.IP_M, mag, err, h_KDE_M, self.N)
        elif kernel == 2:
            IP_gaussian2(self.IP_M, mag, err, h_KDE_M, self.N)

        self.VC1 = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        self.VC2 = <DTYPE_t*>PyMem_Malloc(self.N*sizeof(DTYPE_t))
        if not self.VC1:
            raise MemoryError()
        if not self.VC2:
            raise MemoryError()

        # Precompute terms related to the magnitudes
        self.VM1 = 0.0
        for i in range(self.N):
            self.VC1[i] = 0.0
        for i in range(self.N):
            mat_idx = indexMatrixToVector(i, i, self.N)
            self.VC1[i] += self.IP_M[mat_idx]
            self.VM1 += self.IP_M[mat_idx]
            for j in range(i+1, self.N):
                mat_idx = indexMatrixToVector(i, j, self.N)
                self.VC1[i] += self.IP_M[mat_idx]
                self.VC1[j] += self.IP_M[mat_idx]
                self.VM1 += 2.0*self.IP_M[mat_idx]

    def eval_frequency(self, DTYPE_t freq):
        cdef Py_ssize_t i, j
        for i in range(self.N):
            self.angle[i] = 2.0*M_PI*fmodf(self.mjd[i], 1.0/freq)*freq # output in [0.0, 2.0*pi]
#        IP_wrappednormal(self.IP_P, self.angle, self.h_KDE_P, self.N)
        IP_wrappedcauchy(self.IP_P, self.angle, self.h_KDE_P, self.N)
        cdef Py_ssize_t mat_idx
        cdef DTYPE_t VM1=0.0, VM2=0.0, VC=0.0, VJ=0.0
        for i in range(self.N):
            self.VC2[i] = 0.0    
        for i in range(self.N):
            mat_idx = indexMatrixToVector(i, i, self.N)
            self.VC2[i] += self.IP_P[mat_idx]
            VM2 += self.IP_P[mat_idx]
            VJ += self.IP_M[mat_idx]*self.IP_P[mat_idx]
            for j in range(i+1, self.N):
                mat_idx = indexMatrixToVector(i, j, self.N)
                VM2 += 2.0*self.IP_P[mat_idx]
                self.VC2[j] += self.IP_P[mat_idx]
                self.VC2[i] += self.IP_P[mat_idx]
                VJ += 2.0*self.IP_M[mat_idx]*self.IP_P[mat_idx]
        for i in range(self.N):
            VC += self.VC1[i]*self.VC2[i]
        if self.mode == 0:  # Cauchy-Schwarz MI
            # The log(N) terms cancel out in this sum
            return logf(self.VM1*VM2) + logf(VJ) - 2.0*logf(VC)
        elif self.mode == 1:  # Euclidean MI
            return (self.VM1*VM2/self.N**2 + VJ - 2.0*VC/self.N)/self.N**2
        elif self.mode == 2:  # Quadratic Mutual Entropy, not safe yet
            # Using Renyi's formulation
            return fabsf(-logf(self.VM1) - logf(VM2) + logf(VJ) + 2*logf(self.N))
            # Using Tsallis' formulation
            # return 1.0 - (self.VM1 + VM2 - VJ)/self.N**2  # Tsallis

    def __dealloc__(self):
        PyMem_Free(self.IP_M)
        PyMem_Free(self.IP_P)
        PyMem_Free(self.mjd)
        PyMem_Free(self.angle)
        PyMem_Free(self.VC1)
        PyMem_Free(self.VC2)





import numpy as np


class BasePeriodogram:

    def get_best_frequency(self, fid=None):
        if fid is None:
            best_idx = np.argmax(self.per)
        else:
            best_idx = np.argmax(self.per_single_band[fid])
        return self.freq[best_idx]
        
    def get_best_frequencies(self):
        """
        Returns the best n_local_max frequencies
        """
        
        return self.freq[self.best_local_optima], self.per[self.best_local_optima]
        
    def get_periodogram(self, fid=None):
        if fid is None:
            return self.freq, self.per
        else:
            return self.freq, self.per_single_band[fid]
        
    def frequency_grid_evaluation(self, fmin=0.0, fmax=1.0, fresolution=1e-4):
        """ 
        Computes the selected criterion over a grid of frequencies 
        with limits and resolution specified by the inputs. After that
        the best local maxima are evaluated over a finer frequency grid
        
        Parameters
        ---------
        fmin: float
            starting frequency
        fmax: float
            stopping frequency
        fresolution: float
            step size in the frequency grid
        
        """
        self.freq_step_coarse = fresolution
        freqs = np.arange(start=np.amax([fmin, fresolution]), stop=fmax, 
                          step=fresolution, dtype=np.float32)
        self.per, self.per_single_band = self._compute_periodogram(freqs)
        self.freq = freqs 
    
    def find_local_maxima(self, n_local_optima=10):
        
        local_optima_index = 1+np.where((self.per[1:-1] > self.per[:-2]) & (self.per[1:-1] > self.per[2:]))[0]
        
        if(len(local_optima_index) < n_local_optima):
            print("Warning: Not enough local maxima found in the periodogram")
            return
        # Keep only n_local_optima
        best_local_optima = local_optima_index[np.argsort(self.per[local_optima_index])][::-1]
        if n_local_optima > 0:
            best_local_optima = best_local_optima[:n_local_optima]
        else:
            best_local_optima = best_local_optima[0]
            
        return best_local_optima
    
    def finetune_best_frequencies(self, fresolution=1e-5, n_local_optima=10):
        """
        Computes the selected criterion over a grid of frequencies 
        around a specified amount of  local optima of the periodograms. This
        function is intended for additional fine tuning of the results obtained
        with grid_search
        """
        #assert fresolution < self.freq_step_coarse, "Frequency step error: fine tune step is greater than or equal to coarse step"
        local_optima_index = self.find_local_maxima(n_local_optima)
        
        for local_optimum_index in local_optima_index:
            fmin = self.freq[local_optimum_index] - self.freq_step_coarse
            fmax = self.freq[local_optimum_index] + self.freq_step_coarse
            freqs_fine = np.arange(fmin, fmax, step=fresolution, dtype=np.float32)
            pers_fine = self._compute_periodogram(freqs_fine)
            self._update_periodogram(local_optimum_index, freqs_fine, pers_fine)

        # Sort them in descending order
        idx = np.argsort(self.per[local_optima_index])[::-1]
        if n_local_optima > 0:
            self.best_local_optima = local_optima_index[idx]
        else:
            self.best_local_optima = local_optima_index



#!/usr/bin/python
# -*- coding: utf-8 -*-

from __future__ import division, print_function
import numpy as np
from .base_periodogram import BasePeriodogram
from .algorithms.mutual_information import QMI
from .algorithms.string_length import LKSL
from .algorithms.phase_dispersion_minimization import PDM
from .algorithms.analysis_of_variance import AOV
from .algorithms.multiharmonic_aov import MHAOV
from .math import robust_loc, robust_scale
#from joblib import Parallel, delayed

from collections import namedtuple
Stats = namedtuple('Stats', 'loc scale N')
    
class MultiBandPeriodogram(BasePeriodogram):
    
    def __init__(self, method='MHAOV', **kwarg):
        
        #methods = ["MHAOV"]
        #if not method in methods:
        #    raise ValueError("Wrong method")
        self.method = method
        
        self.Nharmonics = 1
        if 'Nharmonics' in kwarg:
            self.Nharmonics = kwarg["Nharmonics"]
            
        
        
    def set_data(self, mjds, mags, errs, fids):
        
        self.filter_names = np.unique(fids)
        self.mjds = mjds.astype('float32')
        self.mags = mags.astype('float32')
        self.errs = errs.astype('float32')
        self.cython_per = {}
        self.lc_stats = {}
        for filter_name in self.filter_names:
            mask = fids == filter_name
            weights = np.power(self.errs[mask], -2.0)
            weights = weights/np.sum(weights)
            self.lc_stats.update({filter_name : Stats(loc=robust_loc(self.mags[mask], weights),
                                                      scale=robust_scale(self.mags[mask], weights),
                                                      N=len(self.mjds[mask]))})
                             
            self.cython_per.update({filter_name : MHAOV(self.mjds[mask], 
                                                       self.mags[mask]- self.lc_stats[filter_name].loc, 
                                                       self.errs[mask], Nharmonics=1, mode=0)})
            
    def _compute_periodogram(self, freqs):        
        per_single_band = {}
        per_sum = np.zeros_like(freqs) 
        d1 = 2*self.Nharmonics
        #d1 = 2*self.Nharmonics*len(self.filter_names)
        d2_sum = 0.0
        wvar_sum = 0.0
        
        for filter_name in self.filter_names:
            per = np.array([self.cython_per[filter_name].eval_frequency(freq) for freq in freqs], dtype=np.float32)
            d2 = float(self.lc_stats[filter_name].N - 2*self.Nharmonics - 1)  
            per_single_band.update({filter_name : (d2/d1)*per/(self.cython_per[filter_name].wvar-per)})
            #per_single_band.update({filter_name : per})            
            per_sum += per
            #per_sum +=  d1*per*self.cython_per[filter_name].wvar/(d2 + d1*per)            
            wvar_sum += self.cython_per[filter_name].wvar
            d2_sum += d2
            
            
            
        return d2_sum*per_sum/(d1*(wvar_sum - per_sum)), per_single_band
    
    def _update_periodogram(self, replace_idx, freqs_fine, pers_fine):
        new_best = np.argmax(pers_fine[0])
        if pers_fine[0][new_best] > self.per[replace_idx]:
            self.freq[replace_idx] = freqs_fine[new_best] # frequency
            self.per[replace_idx] = pers_fine[0][new_best] # multiband
            for filter_name in self.filter_names: # single band
                self.per_single_band[filter_name][replace_idx] = pers_fine[1][filter_name][new_best]
    
    

class periodogram(BasePeriodogram):
    def __init__(self, method='QMIEU', n_jobs=1, debug=False):
        """
        Class for light curve periodogram computation

        A light curve is a time series of stellar magnitude or flux. Some 
        light curves are periodic, i.e. their brightness change regularly in 
        time (Cepheids, RR Lyrae, Eclipsing Binaries). 
        
        The word periodogram is used in a broad sense to describe a function
        of the period (or frequency) and not the estimator of the power 
        spectral density. 

        Different criteria can be used to assess periodicity. In this package
        we focus on information theoretic based criteria. For more detailes see
       
        J. C. Principe, "Information Theoretic Learning, Renyi's Entropy 
        and Kernel Perspectives", Springer, 2010, Chapters 2 and 10.
        
        
        Parameters
        ---------
        method: {'PDM1', 'LKSL', 'AOV', 'MHAOV', 'QMICS', 'QMIEU'} 
            Method used to perform the fit
            
            PDM: Phase Dispersion Minimization
            LKSL: Lafler Kinman statistic for string length
            AOV: Analysis of Variance Periodo
            MHAOV: Orthogonal multiharmonic AoV
            QMICS: Cauchy Schwarz Quadratic Mutual Information
            QMIEU: Euclidean Quadratic Mutual Information

        n_jobs: positive integer 
            Number of parallel jobs for periodogram computation, NOT IMPLEMENTED
            
        
        """
        self.freq = None
        self.per = None
        self.debug = debug
        
        if not type(n_jobs) is int:
            raise TypeError("Number of jobs must be an integer")
        if n_jobs < 1:
            raise ValueError("Number of jobs must be greater than 0")
        self.n_jobs = n_jobs
        
        methods = ["QMICS", "QMIEU", "QME", "PDM1", "LKSL", "MHAOV", "AOV"]
        if not method in methods:
            raise ValueError("Wrong method")
        self.method = method
        
    def set_data(self, mjd, mag, err, standardize=False, **kwarg):
        """
        Saves the light curve data, where 
        mjd: modified julian date (time instants)
        mag: stellar magnitude 
        err: photometric error
        
        If standardize is True the data is transformed to have zero
        mean and unit standard deviation. Robust 
        estimators of these quantities are used

        TODO: verify that arrays are non empty, non constant, etc
        """
        
        self.mjd = mjd.astype('float32')
        self.mag = mag.astype('float32')
        self.err = err.astype('float32')
        
        # Nan Filter
        na_mask = np.isnan(self.mjd) | np.isnan(self.mag) | np.isnan(self.err)
        if np.sum(na_mask) > 0:
            print(f"Your data contain {np.sum(na_mask)} NaN values, cleaning")
            self.mjd = self.mjd[~na_mask]
            self.mag = self.mag[~na_mask]
            self.err = self.err[~na_mask]
        
        # Standardization
        weights = np.power(err, -2.0)
        weights = weights/np.sum(weights)
        self.lc_stats = {'loc' : robust_loc(mag, weights), 
                         'scale': robust_scale(mag, weights),
                         'N': len(mjd)}        
        
        if standardize:
            self.mag = (self.mag - self.lc_stats['loc'])/self.lc_stats['scale']
            self.err = self.err/self.lc_stats['scale']
        
        
        if self.method in ['QMICS', 'QMIEU', 'QME']:
            
            hm = 0.9*self.lc_stats['N']**(-0.2)
            if not standardize:
                hm = hm*self.lc_stats['scale']
            if 'h_KDE_M' in kwarg:
                hm = hm*kwarg['h_KDE_M']
            hp = 1.0 # How to choose this more appropietly?
            if 'h_KDE_P' in kwarg:
                hp = hp*kwarg['h_KDE_P']
            kernel = 0  # Select the kernel for the magnitudes, 0 is safe
            if 'kernel' in kwarg:
                kernel = kwarg['kernel']
            if self.debug:
                print("Kernel bandwidths: %f , %f" %(hm, hp))
            if self.method == 'QMICS':
                self.cython_per = QMI(self.mjd, self.mag, self.err, hm, hp, 0, kernel)
            elif self.method == 'QMIEU':
                self.cython_per = QMI(self.mjd, self.mag, self.err, hm, hp, 1, kernel)
            else:
                self.cython_per = QMI(self.mjd, self.mag, self.err, hm, hp, 2, kernel)
                
        elif self.method == 'LKSL':  # Lafler-Kinman Minimum String Length
            self.cython_per = LKSL(self.mjd, self.mag, self.err)
        elif self.method == 'PDM1':  # Phase Dispersion Minimization
            Nbins = 8
            if 'Nbins' in kwarg:
                Nbins = kwarg['Nbins']
            self.cython_per = PDM(self.mjd, self.mag, self.err, Nbins)
        elif self.method == 'AOV':  # Analysis of Variance periodogram
            Nbins = 8
            if 'Nbins' in kwarg:
                Nbins = kwarg['Nbins']
            use_errorbars = 1
            if 'use_errorbars' in kwarg:
                use_errorbars = kwarg['use_errorbars']
            self.cython_per = AOV(self.mjd, self.mag, self.err, Nbins, use_errorbars)
        elif self.method == 'MHAOV':  # Orthogonal Multiharmonics AoV periodogram
            Nharmonics = 1
            if 'Nharmonics' in kwarg:
                Nharmonics = kwarg["Nharmonics"]
            self.cython_per = MHAOV(self.mjd, self.mag, self.err, Nharmonics)  

    def _compute_periodogram(self, freqs):
     
        if self.n_jobs == 1:
            pers = np.array([ self.cython_per.eval_frequency(freq) for freq in freqs], dtype=np.float32)            
        #else:
        #    pers = Parallel(n_jobs=self.n_jobs)(delayed(self.compute_metric)(freq) for freq in freqs)

        if self.method in ["PDM1", "LKSL"]: # Minima are best
            pers = -pers
        return pers, None # TODO: THIS IS UGLY!!
    
    def _update_periodogram(self, replace_idx, freqs_fine, pers_fine):
        new_best = np.argmax(pers_fine[0])
        if pers_fine[0][new_best] > self.per[replace_idx]:
            self.freq[replace_idx] = freqs_fine[new_best] 
            self.per[replace_idx] = pers_fine[0][new_best]
           
    